use clap::Parser;
use colored::Colorize;
use memmap2::Mmap;
use onebrc::implementations::rgk::{
    find_city_names, format, run_parallel, to_str, Args, Record, S,
};
use std::thread::available_parallelism;

fn main() {
    let args = Args::parse();

    let start = std::time::Instant::now();
    let filename = args
        .input
        .unwrap_or("../../../measurements.txt".to_string());
    let mmap: Mmap;
    let data;
    {
        let file = std::fs::File::open(filename).unwrap();
        mmap = unsafe { Mmap::map(&file).unwrap() };
        data = &*mmap;
    }

    // Guaranteed to be aligned for SIMD.
    let offset = unsafe { data.align_to::<S>().0.len() };
    let data = &data[offset..];

    // Build a perfect hash function on the cities found in the first 100k characters.
    let names = find_city_names(&data[..4000000]);

    if args.stats {
        eprintln!("Num cities: {}", names.len());
        let mut lens = vec![0; 102];
        for n in &names {
            if *n.last().unwrap() == b';' {
                continue;
            }
            lens[n.len()] += 1;
        }
        for (len, count) in lens.iter().enumerate() {
            if *count != 0 {
                eprintln!("{}: {}", len, count);
            }
        }
    }

    let phf = run_parallel(
        data,
        &names,
        args.threads
            .unwrap_or(available_parallelism().unwrap().into()),
    );

    if args.print {
        print!("{{");
        let mut first = true;

        let mut keys = phf.keys.clone();
        keys.sort_by(|kl, kr| to_str(kl).cmp(to_str(kr)));

        for name in &keys {
            if *name.last().unwrap() != b';' {
                continue;
            }
            let namepos = &name[..name.len() - 1];

            let rpos = phf.index(namepos);
            let rneg = phf.index(name);
            let (min, avg, max) = Record::merge_pos_neg(rpos, rneg);

            if !first {
                print!(", ");
            }
            first = false;

            print!(
                "{}={}/{}/{}",
                to_str(namepos),
                format(min),
                format(avg),
                format(max)
            );
        }
        println!("}}");
    }

    eprintln!(
        "total: {}",
        format!("{:>5.2?}", start.elapsed()).bold().green()
    );
}
